//----------------------------------------------------------------------------
// * Simulation Initialization {simulation_init}
//----------------------------------------------------------------------------
//	Initialize the network with empty data sets, neighbor exclusion, and
//	and a network with a single unrooted node and interleaving scheduling.
//----------------------------------------------------------------------------
Simulation::Simulation(){
	network = new Tree;

	method = Interleaving;
	exclude_neighbors = true;

	tree_heights = new Data_Set<unsigned int, int>("Tree Heights");
	leaf_counts = new Data_Set<unsigned int, int>("Tree Leaf Counts");
	transitions = new Data_Set<unsigned int, int>("State Transitions");
	actions = new Data_Set<unsigned int, int>("Total Actions");
	causality = new Data_Set<unsigned int, int>("Causality Number");
	
	overhead["Before Broadcast"] = new Data_Set<unsigned int, int>("Before Broadcast");
	overhead["After Broadcast"] = new Data_Set<unsigned int, int>("After Broadcast");
	overhead["After Convergecast"] = new Data_Set<unsigned int, int>("After Convergecast");
	overhead["Mixed"] = new Data_Set<unsigned int, int>("Mixed");
}

//----------------------------------------------------------------------------
// * Simulation Initialization {simulation_init_scheduler}
//----------------------------------------------------------------------------
//	Initialize the network with empty data sets, neighbor exclusion, and
//	and a network with a single unrooted node and a given scheduler.
//----------------------------------------------------------------------------
Simulation::Simulation(schedulers daemon){
	network = new Tree;

	method = daemon;
	exclude_neighbors = true;

	tree_heights = new Data_Set<unsigned int, int>("Tree Heights");
	leaf_counts = new Data_Set<unsigned int, int>("Tree Leaf Counts");
	transitions = new Data_Set<unsigned int, int>("State Transitions");
	actions = new Data_Set<unsigned int, int>("Total Actions");
	causality = new Data_Set<unsigned int, int>("Causality Number");
	
	overhead["Before Broadcast"] = new Data_Set<unsigned int, int>("Before Broadcast");
	overhead["After Broadcast"] = new Data_Set<unsigned int, int>("After Broadcast");
	overhead["After Convergecast"] = new Data_Set<unsigned int, int>("After Convergecast");
	overhead["Mixed"] = new Data_Set<unsigned int, int>("Mixed");
}

//----------------------------------------------------------------------------
// * Simulation Copy Constructor {simulation_copy_constructor}
//----------------------------------------------------------------------------
//	Initializes the simulation engine with the copied parameters, data,
//	and network settings of another.
//----------------------------------------------------------------------------
Simulation::Simulation(const Simulation &copy){
	network = new Tree(*copy.network);

	method = copy.method;
	exclude_neighbors = copy.exclude_neighbors;

	tree_heights = new Data_Set<unsigned int, int>(*copy.tree_heights);
	leaf_counts = new Data_Set<unsigned int, int>(*copy.leaf_counts);
	transitions = new Data_Set<unsigned int, int>(*copy.transitions);
	actions = new Data_Set<unsigned int, int>(*copy.actions);
	causality = new Data_Set<unsigned int, int>(*copy.causality);

	std::map<std::string, Data_Set<unsigned int, int>*>::const_iterator overhead_data;

	for(overhead_data = copy.overhead.begin(); overhead_data != copy.overhead.end(); overhead_data++){
		Data_Set<unsigned int, int> *data;
		data = new Data_Set<unsigned int, int>(*overhead_data->second);

		overhead[overhead_data->first] = data;
	}
}

//----------------------------------------------------------------------------
// * Simulation Destructor
//----------------------------------------------------------------------------
//	Deletes the simulation contents, namely the network and data set
//	pointers.
//----------------------------------------------------------------------------
Simulation::~Simulation(){
	delete network;

	delete tree_heights;
	delete leaf_counts;
	delete transitions;
	delete actions;
	delete causality;

	std::map<std::string, Data_Set<unsigned int, int>*>::iterator data;

	for(data = overhead.begin(); data != overhead.end(); data++){
		delete data->second;
	}
}
	
//----------------------------------------------------------------------------
// * Exluding Neighbors? {simulation_excluding} 
//----------------------------------------------------------------------------
//	Returns whether the simulation is currently excluding enabled
//	neighbors during concurrent action execution. 
//----------------------------------------------------------------------------
bool Simulation::excluding() const{
	return exclude_neighbors;
}

//----------------------------------------------------------------------------
// * Get Scheduler {simulation_get_scheduler}
//----------------------------------------------------------------------------
//	Returns the name of the simulation's current scheduler as a string.
//----------------------------------------------------------------------------
std::string Simulation::scheduler() const{
	std::string method_str;

	if(method == Synchronous) method_str = "Synchronous";
	if(method == Power_Set) method_str = "Power Set";
	if(method == Interleaving) method_str = "Interleaving";

	return method_str;
}

//----------------------------------------------------------------------------
// * Set Scheduler {simulation_set_scheduler}
//----------------------------------------------------------------------------
//	Sets the simulaton's scheduler (or daemon) to a new one.
//----------------------------------------------------------------------------
void Simulation::set_method(schedulers daemon){
	method = daemon;
}

//----------------------------------------------------------------------------
// * Set Network Size {simulation_set_size}
//----------------------------------------------------------------------------
//	Sets the network size, and then initializes the network to initiate
//	the size change
//----------------------------------------------------------------------------
void Simulation::set_size(unsigned int net_size){
	network->generate(net_size);
}

//----------------------------------------------------------------------------
// * Reset Data Sets {simulation_reset_data}
//----------------------------------------------------------------------------
//	Clears all data sets leaving them empty for new record keeping
//----------------------------------------------------------------------------
void Simulation::reset_data(){
	tree_heights->reset();
	leaf_counts->reset();
	transitions->reset();
	actions->reset();
	causality->reset();

	overhead["Before Broadcast"]->reset();
	overhead["After Broadcast"]->reset();
	overhead["After Convergecast"]->reset();
	overhead["Mixed"]->reset();
}

//----------------------------------------------------------------------------
// * Toggle Neighbor Exclusion {simulation_toggle_exclude}
//----------------------------------------------------------------------------
//	If simulation is currently exluding enabled neighbors, set the system
//	to ignore such exclusion. If it is not exluding, begin excluding.
//----------------------------------------------------------------------------
void Simulation::toggle_exclude(){
	if(exclude_neighbors){
		exclude_neighbors = false;
	}
	else{
		exclude_neighbors = true;
	}
}

//----------------------------------------------------------------------------
// * Fault Network {simulation_perturb}
//----------------------------------------------------------------------------
//	Randomly inflict a given number of transient faults on the network.
//	Returns the classification of these faults for overhead as a string.
//----------------------------------------------------------------------------
std::string Simulation::perturb(unsigned int to_inflict){
	std::string fault_type;

	if(to_inflict > network->size()){
		to_inflict = network->size();
	}

	std::vector<Node*> available;
	for(std::vector<Node>::iterator node_p = network->begin(); node_p < network->end(); node_p++){
		available.push_back(&(*node_p));
	}

	std::vector<Node*>::iterator pick;
	states original_state;

	while(to_inflict > 0){
		pick = available.begin() + (rand() % available.size());
		original_state = (*pick)->perturb();

		if(fault_type == ""){
			switch(original_state){
			case Idle:
				fault_type = "Before Broadcast";
				break;
			case Requesting:
				fault_type = "After Broadcast";
				break;
			case Replying:
				fault_type = "After Convergecast";
				break;
			}
		}
		else if(fault_type != "Mixed"){
			if((original_state == Idle && fault_type != "Before Broadcast")
			|| (original_state == Requesting && fault_type != "After Broadcast")
			|| (original_state == Replying && fault_type != "After Convergecast")){
				fault_type = "Mixed";
			}
		}

		available.erase(pick);
		to_inflict--;
	}

	return fault_type;
}

//----------------------------------------------------------------------------
// * Get Enabled Nodes {simulation_get_enabled}
//----------------------------------------------------------------------------
//	Returns a vector of node pointers which refer to all nodes which have
//	an action enabled in the current program state. If neighbor exclusion
//	is set to true, the method repeatedly sweeps through all node pairs
//	and if any are found to be linked, one of them is randomly discluded.
//----------------------------------------------------------------------------
std::vector<Node*> Simulation::get_enabled(){
	std::vector<Node*> enabled_nodes;

	for(std::vector<Node>::iterator node = network->begin(); node < network->end(); node++){
		if(node->enabled() > 0){
			enabled_nodes.push_back(&(*node));
		}
	}

	if(excluding()){
		bool found;
		std::vector<Node*>::iterator node;
		std::vector<Node*>::iterator other;

		do{
			found = false;

			for(node = enabled_nodes.begin(); node < enabled_nodes.end(); node++){
				for(other = enabled_nodes.begin(); other < enabled_nodes.end(); other++){
					if((*node)->neighboring(*other)){
						found = true;

						if(rand() % 2 == 0){
							enabled_nodes.erase(node);
						}
						else{
							enabled_nodes.erase(other);
						}
						break;
					}
				}
				if(found) break;
			}
		}while(found);
	}

	return enabled_nodes;
}

//----------------------------------------------------------------------------
// * Get Enabled Nodes {simulation_get_enabled_const}
//----------------------------------------------------------------------------
//	Returns a vector of constant node pointers which refer to all nodes
//	which have an action enabled in the current program state. If neighbor
//	exclusion is set to true, the method repeatedly sweeps through all
//	node pairs and if any are found to be linked, one of them is randomly
//	discluded.
//----------------------------------------------------------------------------
std::vector<const Node*> Simulation::get_enabled() const{
	std::vector<const Node*> enabled_nodes;

	for(std::vector<Node>::const_iterator node = network->begin(); node < network->end(); node++){
		if(node->enabled() > 0){
			enabled_nodes.push_back(&(*node));
		}
	}

	if(excluding()){
		bool found;
		std::vector<const Node*>::iterator node;
		std::vector<const Node*>::iterator other;

		do{
			found = false;

			for(node = enabled_nodes.begin(); node < enabled_nodes.end(); node++){
				for(other = enabled_nodes.begin(); other < enabled_nodes.end(); other++){
					if((*node)->neighboring(*other)){
						found = true;

						if(rand() % 2 == 0){
							enabled_nodes.erase(node);
						}
						else{
							enabled_nodes.erase(other);
						}
						break;
					}
				}
				if(found) break;
			}
		}while(found);
	}

	return enabled_nodes;
}

//----------------------------------------------------------------------------
// * Propagate Network {simulation_propagate}
//----------------------------------------------------------------------------
//	Propagates the system a random number of program state transitions
//	without recording any data.
//----------------------------------------------------------------------------
void Simulation::propagate(){
	unsigned int steps = rand() % 1001;

	while(steps > 0){
		std::vector<Node*> enabled = get_enabled();
		std::vector<Node*>::iterator acting;
		unsigned int to_execute;

		switch(method){
		case Synchronous:
			for(acting = enabled.begin(); acting < enabled.end(); acting++){
				(*acting)->execute();
			}
			break;
		case Power_Set:
			to_execute = rand() % enabled.size() + 1;
			while(to_execute > 0){
				acting = enabled.begin() + (rand() % enabled.size());
				(*acting)->execute();
				enabled.erase(acting);

				to_execute--;
			}
			break;
		case Interleaving:
			acting = enabled.begin() + (rand() % enabled.size());
			(*acting)->execute();
			break;
		}
		steps--;
	}
}

//----------------------------------------------------------------------------
// * Run algorithm Computation {simulation_compute}
//----------------------------------------------------------------------------
//	Inflicts a number of transient faults on a self-stabilizing algorithm,
//	and then runs the algorithm until it has reached a legal program
//	state. Returns the number of state transitions before reaching a legal
//	state. Returns zero if the program was found to remain in a legal
//	program state after fault occurence.
//
//	NOTE: This method updates the overhead and actions counter data sets.
//----------------------------------------------------------------------------
int Simulation::compute(unsigned int faults, bool scale_test){
	std::string overhead_type = perturb(faults);

	int overhead_actions = 0;
	int state = 0;

	while(!network->legitimate()){
		std::vector<Node*> enabled = get_enabled();
		std::vector<Node*>::iterator acting;
		unsigned int to_execute;

		switch(method){
		case Synchronous:
			for(acting = enabled.begin(); acting < enabled.end(); acting++){
				(*acting)->execute();
				(*acting)->update_causality();
				if(network->overhead(*acting)) overhead_actions++;
			}
			break;
		case Power_Set:
			to_execute = rand() % enabled.size() + 1;
			while(to_execute > 0){
				acting = enabled.begin() + (rand() % enabled.size());
				(*acting)->execute();
				(*acting)->update_causality();
				if(network->overhead(*acting)) overhead_actions++;

				enabled.erase(acting);
				to_execute--;
			}
			break;
		case Interleaving:
			acting = enabled.begin() + (rand() % enabled.size());
			(*acting)->execute();
			(*acting)->update_causality();
			if(network->overhead(*acting)) overhead_actions++;

			break;
		}
		state++;
	}

	if(scale_test){
		causality->add_point(network->size(), network->max_causality());
		overhead[overhead_type]->add_point(network->size(), overhead_actions);
	}
	else{
		causality->add_point(faults, network->max_causality());
		overhead[overhead_type]->add_point(faults, overhead_actions);
	}	

	return state;
}

//----------------------------------------------------------------------------
// * Run Fault Sweep {simulation_run_comps_bounds}
//----------------------------------------------------------------------------
//	Runs a transient fault sweep test on the algorithm by repetitively
//	initializing the network to a random legitimate state, inflicting an
//	incremental number of transient faults ranging between the lower and
//	upper bounds, and invoking a practical computation. Each computation
//	is repeated comps many times before fault incrementation. All data
//	metrics are collected in the various Data Set objects belonging to
//	this simulation engine.
//----------------------------------------------------------------------------
void Simulation::run(unsigned int comp_limit, unsigned int lower_bound, unsigned int upper_bound){
	unsigned int computation;
	unsigned int faults;
	int state_count;

	for(faults = lower_bound; faults <= upper_bound; faults++){
		computation = 0;

		while(computation < comp_limit){
			network->generate(network->size());
			propagate();
			network->reset_actions();

			state_count = compute(faults);

			tree_heights->add_point(network->size(), network->height());
			leaf_counts->add_point(network->size(), network->leaf_count());
			transitions->add_point(faults, state_count);
			actions->add_point(faults, network->total_actions());

			computation++;
		}
	}
}		

//----------------------------------------------------------------------------
// * Run Fault Sweep {simulation_run_bounds}
//----------------------------------------------------------------------------
//	Runs a fault sweep test at a default number of computations per
//	transient fault ranging from lower bound to upper bound.
//----------------------------------------------------------------------------
void Simulation::run(unsigned int lower_bound, unsigned int upper_bound){
	run(COMPS, lower_bound, upper_bound);
}

//----------------------------------------------------------------------------
// * Run Fault Sweep {simulation_run_comps}
//----------------------------------------------------------------------------
//	Runs a fault sweep test with transient faults ranging from 1 to the
//	total system size at a given number of computations per fault.
//----------------------------------------------------------------------------
void Simulation::run(unsigned int comp_limit){
	run(comp_limit, 1, network->size());
}

//----------------------------------------------------------------------------
// * Run Fault Sweep {simulation_run_default}
//----------------------------------------------------------------------------
//	Runs a fault sweet test at a default number of computations per
//	transient fault ranging from 1 to the total system size.
//----------------------------------------------------------------------------
void Simulation::run(){
	run(COMPS, 1, network->size());
}

//----------------------------------------------------------------------------
// * Run Scale Test {simulation_run_scale_bounds_comps}
//----------------------------------------------------------------------------
//	Runs a scale test with network size increased from lower_bound to
//	upper_bound by a given step value, introducing a number of transient
//	faults set by a constant rate (with regards to network size), running 
//	comp_limit number of computations for each size. The results of this
//	experiment are stored in the simulation data objects with system size
//	used as a key.
//----------------------------------------------------------------------------
void Simulation::run_scale(unsigned int fault_rate, unsigned int lower_bound, unsigned int upper_bound, unsigned int step, unsigned int comp_limit){
	unsigned int system_size;
	unsigned int computation;
	int state_count;

	if(fault_rate > 100){
		std::cerr << "Error: illegal fault rate used for scale test. Must be a percent rate equal to or less than 100" << std::endl;
	}
	else{
		for(system_size = lower_bound; system_size <= upper_bound; system_size += step){
			computation = 0;

			while(computation < comp_limit){
				network->generate(system_size);
				propagate();
				network->reset_actions();

				state_count = compute((fault_rate * system_size) / 100, true);
	
				tree_heights->add_point(network->size(), network->height());
				leaf_counts->add_point(network->size(), network->leaf_count());
				transitions->add_point(network->size(), state_count);
				actions->add_point(network->size(), network->total_actions());

				computation++;
			}
		}
	}
}	

//----------------------------------------------------------------------------
// * Run Scale Test {simulation_run_scale_bounds}
//----------------------------------------------------------------------------
//	Runs a scale test with network size increased from lower_bound to
//	upper_bound by a given step value, introducing a number of transient
//	faults set by a constant rate (with regards to network size), running 
//	the default number of computations for each size. The results of this
//	experiment are stored in the simulation data objects with system size
//	used as a key.
//----------------------------------------------------------------------------
void Simulation::run_scale(unsigned int fault_rate, unsigned int lower_bound, unsigned int upper_bound, unsigned int step){
	run_scale(fault_rate, lower_bound, upper_bound, step, COMPS);
}

//----------------------------------------------------------------------------
// * Run Scale Test {simulation_run_scale_default}
//----------------------------------------------------------------------------
//	Runs a scale test with network size increased from 100 to
//	1000 nodes in steps of 100, introducing a number of transient faults
//	set by a constant rate (with regards to network size), running the
//	default number of computations for each size. The results of this
//	experiment are stored in the simulation data objects with system size
//	used as a key.
//----------------------------------------------------------------------------
void Simulation::run_scale(unsigned int fault_rate){
	run_scale(fault_rate, 100, 1000, 100, COMPS);
}

//----------------------------------------------------------------------------
// * Print Test Data {simulation_print_results}
//----------------------------------------------------------------------------
//	Prints the data contained in all measurement data sets.
//----------------------------------------------------------------------------
void Simulation::print_results(std::ostream &out, bool header) const{
	unsigned int fault;

	out << "Tree Height: " << tree_heights->mean(100) << std::endl;
	out << "Leaf Count: " << leaf_counts->mean(100) << std::endl;
	out << std::endl;

	if(header){
		out << "Faults\tTransitions\tActions\tCausality\tBBcast\tABcast\tACcast\tMixed" << std::endl;
	}


	for(fault = 1; fault <= 100; fault++){
		out << fault;
		out << '\t';
		out << transitions->mean(fault);
		out << "\t\t";
		out << actions->mean(fault);
		out << '\t';
		out << causality->mean(fault);
		out << '\t';

		std::map<std::string, Data_Set<unsigned int, int>*>::const_iterator overhead_data;

		for(overhead_data = overhead.begin(); overhead_data != overhead.end(); overhead_data++){
			out << '\t';
			out << overhead_data->second->mean(fault);
		}
		out << std::endl;
	}
	out << std::endl;
}

void Simulation::print_results(bool header) const{
	print_results(std::cout, header);
}
