//----------------------------------------------------------------------------
// * Tree Default Initialization {tree_init_default}
//----------------------------------------------------------------------------
//	Initializes the tree as a single unrooted node.
//----------------------------------------------------------------------------
Tree::Tree(){
	data_root = 0;

	Node node(1);
	nodes.push_back(node);
}

//----------------------------------------------------------------------------
// * Tree Initialization {tree_init_size}
//----------------------------------------------------------------------------
//	Initialize tree with user defined tree size.
//----------------------------------------------------------------------------
Tree::Tree(unsigned int tree_size){
	generate(tree_size);
}

//----------------------------------------------------------------------------
// * Tree Initialization {tree_init_prufer}
//----------------------------------------------------------------------------
//	Initialize tree with user defined prufer sequence.
//----------------------------------------------------------------------------
Tree::Tree(const Prufer &prufer){
	generate(prufer);
}

//----------------------------------------------------------------------------
// * Tree Copy Construction {tree_copy_constructor}
//----------------------------------------------------------------------------
//	Initialize tree with the copied contents of another, reassigning link
//	poiners, the root node pointer, and hierarchy pointers.
//----------------------------------------------------------------------------
Tree::Tree(const Tree &copy){
	std::vector<Node>::iterator node;
	std::vector<Node>::const_iterator copy_node;
	std::vector<const Node*>::const_iterator copy_link;

	for(copy_node = copy.begin(); copy_node < copy.end(); copy_node++){
		Node new_node = *copy_node;
		nodes.push_back(new_node);
	}

	for(node = nodes.begin(), copy_node = copy.begin(); node < nodes.end(); node++, copy_node++){
		for(copy_link = copy_node->links_begin(); copy_link < copy_node->links_end(); copy_link++){
			node->add_link(this->operator[]((*copy_link)->id()));
		}
	}

	data_root = this->operator[](copy.root()->id());
	create_hierarchy(data_root);
}

//----------------------------------------------------------------------------
// * Generate Default Tree {tree_generate_default} 
//----------------------------------------------------------------------------
//	Generates a random tree of 100 nodes from a random prufer sequence.
//	Sets a random root and legitimate initial state.
//----------------------------------------------------------------------------
void Tree::generate(){
	Prufer prufer(98);

	decode(prufer);
	data_root = &(nodes[rand() % size()]);
	create_hierarchy(root());
	legitimize(root());
	define_actions();
}

//----------------------------------------------------------------------------
// * Generate Tree {tree_generate_size}
//----------------------------------------------------------------------------
//	Generates a random tree of tree_size nodes using a random prufer
//	sequence. Random root and legitimate initial state.
//----------------------------------------------------------------------------
void Tree::generate(unsigned int tree_size){
	Prufer prufer(tree_size - 2);

	decode(prufer);
	data_root = &(nodes[rand() % size()]);
	create_hierarchy(root());
	legitimize(root());
	define_actions();
}

//----------------------------------------------------------------------------
// * Generate Tree {tree_generate_prufer}
//----------------------------------------------------------------------------
//	Generates a tree using a user defined prufer sequence. Random root and
//	legitimate initial state.
//----------------------------------------------------------------------------
void Tree::generate(const Prufer &prufer){
	decode(prufer);
	data_root = &(nodes[rand() % size()]);
	create_hierarchy(root());
	legitimize(root());
	define_actions();
}

//----------------------------------------------------------------------------
// * Generate Default Unrooted Tree {tree_generate_unrooted_default}
//----------------------------------------------------------------------------
//	Generates a random unrooted tree of 100 nodes using a random prufer
//	sequence.
//----------------------------------------------------------------------------
void Tree::generate_unrooted(){
	Prufer prufer(98);

	decode(prufer);
}

//----------------------------------------------------------------------------
// * Generate Unrooted Tree {tree_generate_unrooted_size}
//----------------------------------------------------------------------------
//	Generates a random unrooted tree of tree_size nodes using a random
//	prufer sequence.
//----------------------------------------------------------------------------
void Tree::generate_unrooted(unsigned int tree_size){
	Prufer prufer(tree_size - 2);

	decode(prufer);
}

//----------------------------------------------------------------------------
// * Generate Unrooted Tree {tree_generate_unrooted_prufer}
//----------------------------------------------------------------------------
//	Generates an unrooted tree using a user defined prufer sequence.
//----------------------------------------------------------------------------
void Tree::generate_unrooted(const Prufer &prufer){
	decode(prufer);
}

//----------------------------------------------------------------------------
// * Get Tree Size {tree_get_size}
//----------------------------------------------------------------------------
//	Returns the tree's size, or the total number of nodes in the tree.
//----------------------------------------------------------------------------
unsigned int Tree::size() const{
	return nodes.size();
}

//----------------------------------------------------------------------------
// * Tree Empty? {tree_empty}
//----------------------------------------------------------------------------
//	Returns true if the tree has no nodes, and false otherwise.
//----------------------------------------------------------------------------
bool Tree::empty() const{
	return nodes.empty();
}

//----------------------------------------------------------------------------
// * Tree is Rooted? {tree_rooted}
//----------------------------------------------------------------------------
//	Returns true if the tree has a root, returns false otherwise.
//----------------------------------------------------------------------------
bool Tree::rooted() const{
	return(data_root > 0);
}
	
//----------------------------------------------------------------------------
// * Get Tree Root ID {tree_root}
//----------------------------------------------------------------------------
//	Returns a pointer to the tree's root node.
//----------------------------------------------------------------------------
Node * Tree::root(){
	return data_root;
}

//----------------------------------------------------------------------------
// * Get Tree Root ID {tree_root_const}
//----------------------------------------------------------------------------
//	Returns a constant pointer to the tree's root node.
//----------------------------------------------------------------------------
const Node * Tree::root() const{
	return data_root;
}

//----------------------------------------------------------------------------
// * Get Tree Height {tree_height}
//----------------------------------------------------------------------------
//	Returns the height of tree, or the maximum depth of any leaf.
//----------------------------------------------------------------------------
unsigned int Tree::height() const{
	unsigned int max_height = 0;

	const Node *traverser;
	std::vector<Node>::const_iterator node;
	std::deque<const Node*> subchain;

	for(node = nodes.begin(); node < nodes.end(); node++){
		if(node->leaf()){
			traverser = &(*node);

			subchain.push_front(traverser);
			do{
				traverser = traverser->parent();
				subchain.push_front(traverser);
			}while(!traverser->root());

			if(max_height < subchain.size()) max_height = subchain.size();

			subchain.clear();
		}
	}

	return max_height;
}

//----------------------------------------------------------------------------
// * Leaf Count {tree_leaf_count}
//----------------------------------------------------------------------------
//	Returns the total number of leaf processes in the tree.
//----------------------------------------------------------------------------
unsigned int Tree::leaf_count() const{
	unsigned int count = 0;

	for(std::vector<Node>::const_iterator node = nodes.begin(); node < nodes.end(); node++){
		if(node->leaf()) count++;
	}

	return count;
}


//----------------------------------------------------------------------------
// * Get Highest Causality Number {tree_max_causality}
//----------------------------------------------------------------------------
//	Returns the highest causality number between all nodes in the tree.
//----------------------------------------------------------------------------
int Tree::max_causality() const{
	int maximum = 0;
	std::vector<Node>::const_iterator node;

	for(node = nodes.begin(); node < nodes.end(); node++){
		if(node->get_causality() > maximum){
			maximum = node->get_causality();
		}
	}

	return maximum;
}

//----------------------------------------------------------------------------
// * Get Total Actions Performed {tree_total_actions}
//----------------------------------------------------------------------------
//	Returns the total number of actions performed by all nodes in the
//	tree.
//----------------------------------------------------------------------------
int Tree::total_actions() const{
	int total = 0;

	std::vector<Node>::const_iterator node;

	for(node = nodes.begin(); node < nodes.end(); node++){
		total += node->actions_performed();
	}

	return total;
}

//----------------------------------------------------------------------------
// * Calculate if Overhead {tree_overhead}
//----------------------------------------------------------------------------
//	If the active node is part of a faulted subchain, returns true.
//	Returns false otherwise, and also if the tree is not rooted.
//----------------------------------------------------------------------------
bool Tree::overhead(const Node* node) const{
	bool result = false;

	if(rooted()){
		const Node *traverser;
		std::deque<const Node*> subchain;
	
		for(std::vector<const Node*>::const_iterator child = node->children_begin(); child < node->children_end(); child++){
			traverser = *child;

			subchain.push_front(traverser);
			do{
				traverser = traverser->parent();
				subchain.push_front(traverser);
			}while(!traverser->root());

			if(!(RP(subchain) || RQ(subchain))){
				result = true;
				break;
			}

			subchain.clear();
		}
	}

	return result;
}

//----------------------------------------------------------------------------
// * Determine Invariant {tree_legitimate}
//----------------------------------------------------------------------------
//	Returns true if the invariant - RP(k) OR RQ(l, m) evaluates as true
//	for all tree subchains. Returns false otherwise, and also if the tree
//	is not rooted..
//----------------------------------------------------------------------------
bool Tree::legitimate() const{
	bool evaluation = true;

	if(rooted()){
		const Node *traverser;

		std::vector<Node>::const_iterator node;
		std::deque<const Node*> subchain;

		for(node = nodes.begin(); node < nodes.end(); node++){
			if(node->leaf()){
				traverser = &(*node);

				subchain.push_front(traverser);
				do{
					traverser = traverser->parent();
					subchain.push_front(traverser);
				}while(!traverser->root());

				if(!(RP(subchain) || RQ(subchain))){
					evaluation = false;
					break;
				}

				subchain.clear();
			}
		}
	}
	else{
		evaluation = false;
	}

	return evaluation;
}

//----------------------------------------------------------------------------
// * Nodes Begin {tree_begin)
//----------------------------------------------------------------------------
//	Returns the iterator pointing to the first node in the tree.
//----------------------------------------------------------------------------
std::vector<Node>::iterator Tree::begin(){
	return nodes.begin();
}

//----------------------------------------------------------------------------
// * Nodes Begin {tree_begin_const)
//----------------------------------------------------------------------------
//	Returns the constant iterator pointing to the first node in the tree.
//----------------------------------------------------------------------------
std::vector<Node>::const_iterator Tree::begin() const{
	return nodes.begin();
}

//----------------------------------------------------------------------------
// * Nodes End {tree_end)
//----------------------------------------------------------------------------
//	Returns the iterator pointing to the end of the tree.
//----------------------------------------------------------------------------
std::vector<Node>::iterator Tree::end(){
	return nodes.end();
}

//----------------------------------------------------------------------------
// * Nodes End {tree_end_const)
//----------------------------------------------------------------------------
//	Returns the constant iterator pointing to the end of the tree.
//----------------------------------------------------------------------------
std::vector<Node>::const_iterator Tree::end() const{
	return nodes.end();
}

//----------------------------------------------------------------------------
// * Stop Program {tree_stop}
//----------------------------------------------------------------------------
//	Sets all nodes to a stationary state.
//----------------------------------------------------------------------------
void Tree::stop(){
	for(std::vector<Node>::iterator node = nodes.begin(); node < nodes.end(); node++){
		node->set_state(Idle);
	} 
}

//----------------------------------------------------------------------------
// * Reset Action Counters {tree_reset_actions}
//----------------------------------------------------------------------------
//	Resets all node action counters to zero.
//----------------------------------------------------------------------------
void Tree::reset_actions(){
	for(std::vector<Node>::iterator node = nodes.begin(); node < nodes.end(); node++){
		node->reset_actions();
	}
}	

//----------------------------------------------------------------------------
// * Set Tree Root {tree_set_root}
//----------------------------------------------------------------------------
//	Sets the tree root to node with ID = node_id and then defines its
//	descendants. Returns whether a node with this ID was found.
//----------------------------------------------------------------------------
bool Tree::set_root(int node_id){
	bool success = false;

	for(std::vector<Node>::iterator node = nodes.begin(); node < nodes.end(); node++){
		if(node->id() == node_id){
			if(rooted()) unroot();

			data_root = &(*node);
			create_hierarchy(root());
			legitimize(root());
			define_actions();

			success = true;
		}
	}

	return success;
}

//----------------------------------------------------------------------------
// * Set Tree Root {tree_set_root_pointer}
//----------------------------------------------------------------------------
//	Sets the tree root to node and then defines its descendants.
//----------------------------------------------------------------------------
void Tree::set_root(Node* node){
	if(rooted()) unroot();

	data_root = node;
	create_hierarchy(root());
	legitimize(root());
	define_actions();
}

//----------------------------------------------------------------------------
// * Unroot Tree {tree_unroot}
//----------------------------------------------------------------------------
//	Sets the tree root to a non-node and clears all descent/ancestry
//	data.
//----------------------------------------------------------------------------
void Tree::unroot(){
	data_root = 0;

	std::vector<Node>::iterator node;

	for(node = nodes.begin(); node < nodes.end(); node++){
		node->set_parent(0);
		node->disown();
		node->define_actions();
	}
}

//----------------------------------------------------------------------------
// * Tree Subscript Operator {tree_subscript_operator}
//----------------------------------------------------------------------------
//	Returns a pointer to a node in the tree with ID = node_id.
//----------------------------------------------------------------------------
Node* Tree::operator[](int node_id){
	Node *match = 0;

	std::vector<Node>::iterator node;

	for(node = nodes.begin(); node < nodes.end(); node++){
		if(node->id() == node_id){
			match = &(*node);
		}
	}

	return match;
}

//----------------------------------------------------------------------------
// * Tree Subscript Operator {tree_subscript_operator_const}
//----------------------------------------------------------------------------
//	Returns a constant pointer to a node in the tree with ID = node_id.
//----------------------------------------------------------------------------
const Node* Tree::operator[](int node_id) const{
	const Node *match = 0;

	std::vector<Node>::const_iterator node;

	for(node = nodes.begin(); node < nodes.end(); node++){
		if(node->id() == node_id){
			match = &(*node);
		}
	}

	return match;
}

//----------------------------------------------------------------------------
// * Tree Assignment Operator {tree_assignment_operator}
//----------------------------------------------------------------------------
//	Assigns tree with the copied contents of another, reassigning link
//	poiners, the root node pointer, and hierarchy pointers. Returns a
//	pointer to this tree for compound assignment.
//----------------------------------------------------------------------------
Tree* Tree::operator=(const Tree &copy){
	nodes.clear();

	std::vector<Node>::iterator node;
	std::vector<Node>::const_iterator copy_node;
	std::vector<const Node*>::const_iterator copy_link;

	for(copy_node = copy.begin(); copy_node < copy.end(); copy_node++){
		Node new_node = *copy_node;
		nodes.push_back(new_node);
	}

	for(node = nodes.begin(), copy_node = copy.begin(); node < nodes.end(); node++, copy_node++){
		for(copy_link = copy_node->links_begin(); copy_link < copy_node->links_end(); copy_link++){
			node->add_link(this->operator[]((*copy_link)->id()));
		}
	}

	data_root = this->operator[](copy.root()->id());
	create_hierarchy(data_root);

	return this;
}

//----------------------------------------------------------------------------
// * Decode Prufer Sequence {tree_decode}
//----------------------------------------------------------------------------
//	Sets the tree shape to that defined by a decoded prufer sequence.
//---------------------------------------------------------------------------- 
void Tree::decode(const Prufer &prufer){
	nodes.clear();

	std::vector<int> degree(prufer.size() + 2, 1);

	for(int i = 0; i < prufer.size() + 2; i++){
		Node node(i + 1);
		nodes.push_back(node);
	}

	for(int i = 0; i < prufer.size(); i++){
		degree[prufer[i]]++;
	}

	for(int i = 0; i < prufer.size(); i++){
		for(int j = 0; j < degree.size(); j++){
			if(degree[j] == 1){
				add_edge(prufer[i], j);
				degree[prufer[i]]--;
				degree[j]--;
				break;
			}
		}
	}
	
	int u(0), v(0);

	for(int i = 0; i < degree.size(); i++){
		if(degree[i] == 1){
			if(u == 0){
				u = i;
			}
			else{
				v = i;
			}
		}
	}
	add_edge(u, v);
}

//---------------------------------------------------------------------------- 
// * Add Edge {tree_add_edge}
//---------------------------------------------------------------------------- 
//	Draws a bridge between the nodes with node IDs i and j.
//---------------------------------------------------------------------------- 
void Tree::add_edge(int i, int j){
	nodes[i].add_link(&(nodes[j]));
	nodes[j].add_link(&(nodes[i]));
}

//---------------------------------------------------------------------------- 
// * Add Edge {tree_add_edge_pointer}
//---------------------------------------------------------------------------- 
//	Draws a bridge between two nodes.
//---------------------------------------------------------------------------- 
void Tree::add_edge(Node* left, Node* right){
	left->add_link(right);
	right->add_link(left);
}

//---------------------------------------------------------------------------- 
// * Create Tree Hierarchy {tree_create_hierarchy}
//---------------------------------------------------------------------------- 
//	Recursively defines the descendants of a node.
//---------------------------------------------------------------------------- 
void Tree::create_hierarchy(Node* node){
	std::vector<Node*>::iterator link;

	for(link = node->links_begin(); link < node->links_end(); link++){
		if(*link != node->parent()){
			node->add_child(*link);
			(*link)->set_parent(node);
			create_hierarchy(*link);
		}
	}
}

//---------------------------------------------------------------------------- 
// * Legitimize Tree State {tree_legitimize}
//---------------------------------------------------------------------------- 
//	Sets the tree to a random legitimate state as defined by the PIF
//	protocol. Returns false if the tree is unrooted, in which case the
//	method is not defined properly.
//---------------------------------------------------------------------------- 
bool Tree::legitimize(Node* node){
	bool success = true;

	if(rooted()){
		std::vector<states> available;

		if(node->root()){
			available.push_back(Requesting);
			available.push_back(Idle);
		}
		else if(node->intermediate()){
			if(node->parent()->requesting()){
				available.push_back(Requesting);
			}
			if(!node->parent()->replying()){
				available.push_back(Idle);
			}
			available.push_back(Replying);
		}
		else if(node->leaf()){
			if(!node->parent()->replying()){
				available.push_back(Idle);
			}
			available.push_back(Replying);
		}

		node->set_state(available[rand() % available.size()]);

		for(std::vector<Node*>::iterator child = node->children_begin(); child < node->children_end(); child++){
			legitimize(*child);
		}
	}
	else{
		std::cerr << "ERROR: called 'legitimize' on an unrooted tree. 'legitimize' will not be executed." << std::endl;
		success = false;
	}

	return success;
}

//---------------------------------------------------------------------------- 
// * Define Node Actions {tree_define_actions}
//---------------------------------------------------------------------------- 
//	Defines the actions associated with each node in the tree.
//----------------------------------------------------------------------------
void Tree::define_actions(){
	std::vector<Node>::iterator node;

	for(node = nodes.begin(); node < nodes.end(); node++){
		node->define_actions();
	}
}
 
// * Predicate RP(k) {tree_RP(k)}
//---------------------------------------------------------------------------- 
//	Evaluates the predicate RP(k) as defined by the PIF protocol for a
//	given PIF node chain, contained as a double-ended que, for all k
//	integers between 0 and chain size - 1.
//---------------------------------------------------------------------------- 
bool Tree::RP(const std::deque<const Node*> &chain) const{
	bool evaluation = false;

/*	std::cout << "Node chain: {";
	for(int i = 0; i < chain.size(); i++){
		if(chain[i]->requesting()) std::cout << "rq";
		if(chain[i]->idle()) std::cout << "i";
		if(chain[i]->replying()) std::cout << "rp";
		if(i < chain.size() - 1) std::cout << ", ";
	}
	std::cout << "} in RP" << std::endl;
*/
	for(int k = 0; k < chain.size() - 1; k++){
		if(RP(chain, k)){
			evaluation = true;
			break;
		}
	}

	return evaluation;
}

//---------------------------------------------------------------------------- 
// * Predicate RP(k) {tree_RP_single}
//---------------------------------------------------------------------------- 
//	Evaluates the predicate RP(k) as defined by the PIF protocol for a
//	given PIF node chain, contained as a double-ended que, for integer k.
//---------------------------------------------------------------------------- 
bool Tree::RP(const std::deque<const Node*> &chain, int k) const{
	bool evaluation = true;

	for(int i = 0; i < chain.size(); i++){
		if((i <= k && !chain[i]->requesting()) || (i > k && !chain[i]->replying())){
			evaluation = false;
			break;
		}
	}

	return evaluation;
}

//---------------------------------------------------------------------------- 
// * Predicate RQ(l, m) {tree_RQ(l, m)}
//---------------------------------------------------------------------------- 
//	Evaluates the predicate RQ(l, m) as defined by the PIF protocol for a
//	given PIF node chain, contained as a double-ended que, for all l
//	integers between 0 and chain size - 1, and all m integers between
//	l + 1 and chain size.
//---------------------------------------------------------------------------- 
bool Tree::RQ(const std::deque<const Node*> &chain) const{
	bool evaluation = false;

/*	std::cout << "Node chain: {";
	for(int i = 0; i < chain.size(); i++){
		if(chain[i]->requesting()) std::cout << "rq";
		if(chain[i]->idle()) std::cout << "i";
		if(chain[i]->replying()) std::cout << "rp";
		if(i < chain.size() - 1) std::cout << ", ";
	}
	std::cout << "} in RQ" << std::endl;
*/
	for(int l = -1; l < static_cast<int>(chain.size() - 1); l++){
		for(int m = l + 1; m < chain.size(); m++){
			if(RQ(chain, l, m)){
				evaluation = true;
				break;
			}
		}
	}

	return evaluation;
}

//---------------------------------------------------------------------------- 
// * Predicate RQ(l, m) {tree_RQ_single}
//---------------------------------------------------------------------------- 
//	Evaluates the predicate RQ(l, m) as defined by the PIF protocol for a
//	given PIF node chain, contained as a double-ended que, for integers l
//	and m.
//---------------------------------------------------------------------------- 
bool Tree::RQ(const std::deque<const Node*> &chain, int l, int m) const{
	bool evaluation = true;

	for(int i = 0; i < chain.size(); i++){
		if((i <= l && !chain[i]->requesting()) || 
				(l < i && i <= m && !chain[i]->idle()) ||
				(m < i && !chain[i]->replying())){

			evaluation = false;
			break;
		}
	}

	return evaluation;
}

//---------------------------------------------------------------------------- 
// * Default Print {tree_print_default}
//---------------------------------------------------------------------------- 
//	Prints tree information to standard output.
//---------------------------------------------------------------------------- 
void Tree::print() const{
	print(std::cout);
}

//---------------------------------------------------------------------------- 
// * Print to Stream
//---------------------------------------------------------------------------- 
//	Prints tree information to a user defined output.
//---------------------------------------------------------------------------- 
void Tree::print(std::ostream &out) const{
	out << "ID\tType\t\tState\t\tEnabled\tActions\tCausality\tParent\tChildren\t\t\tLinks" << std::endl;

	for(int i = 0; i < 104; i++){
		out << '-';
	}

	out << std::endl;

	std::vector<Node>::const_iterator node;
	for(node = nodes.begin(); node < nodes.end(); node++){
		out << node->id();
		out << '\t';

		if(node->root()) out << "Root\t";
		if(node->intermediate()) out << "Intermediate";
		if(node->leaf()) out << "Leaf\t";
		if(!node->rooted()) out << "N/A\t";
		out << '\t';

		out << node->state_str();
		if(node->idle()) out << '\t';
		out << '\t';

		if(node->enabled() > 0){
			out << node->enabled()->name();
		}
		else{
			out << "n/a";
		}
		out << '\t';

		out << node->actions_performed();
		out << '\t';

		out << node->get_causality();
		out << "\t\t";

		if(!node->root() && node->rooted()){
			out << node->parent()->id();
		}
		else{
			out << "None";
		}
		out << '\t';


		if(!node->rooted() || node->leaf()){
			out << "None\t";
		}
		else{
			for(std::vector<const Node*>::const_iterator child = node->children_begin(); child < node->children_end(); child++){
				std::cout << (*child)->id();
				if(child < node->children_end() - 1){
					std::cout << ", ";
				}
			}
		}
		out << "\t\t";
		
		if(!node->linked()){
			out << "None\t";
		}
		else{
			for(std::vector<const Node*>::const_iterator link = node->links_begin(); link < node->links_end(); link++){
				std::cout << (*link)->id();
				if(link < node->links_end() - 1){
					std::cout << ", ";
				}
			}
		}
		out << '\t';

	
		out << std::endl;
	}

	for(int i = 0; i < 104; i++){
		out << "-";
	}
	out << std::endl;
}

std::ostream& operator<<(std::ostream &out, const Tree &tree){
	tree.print(out);
	return out;
}
